from __future__ import annotations

import dataclasses
import time
import typing as typ
import warnings

import jijmodeling as jm
import numpy as np

from jijmodeling import Problem

from jijzept.client import JijZeptClient
from jijzept.config.path_type import PATH_TYPE
from jijzept.entity.schema import SolverType
from jijzept.exception.exception import (
    JijZeptClientValidationError,
    JijZeptSolvingFailedError,
    JijZeptSolvingUnknownError,
    JijZeptSolvingValidationError,
)
from jijzept.instance_translator.instance_translator import InstanceTranslator
from jijzept.post_api import post_instance_and_query
from jijzept.response import JijModelingResponse
from jijzept.response.base import APIStatus
from jijzept.utils import serialize_fixed_var
from google.protobuf.text_encoding import CEscape

# This class provides only a constructor,
# so each sampler should implement its own methods such as .sample_model.
# In addition, each sampler specifies the API to be connected by specifying "jijmodeling_solver_type",
# so be sure to implement it.
class JijZeptBaseSampler:
    """Parent class for each sampler in JijZept.

    This class only provide the constructor.
    The actual sampling method are implemented by each sampler.
    """

    jijmodeling_solver_type: SolverType

    def __init__(
        self,
        token: str | None = None,
        url: str | None = None,
        proxy: str | None = None,
        config: PATH_TYPE | None = None,
        config_env: str = "default",
    ):
        """Sets token and url.

        If you do not set any arguments, JijZept configuration file is used.
        If you set the url or token here, that will be used as the priority setting for connecting to the API.
        See JijZeptClient for details.

        Args:
            token (str | None, optional): Token string. Defaults to None.
            url (str | None, optional): API URL. Defaults to None.
            proxy (str | None, optional): Proxy URL. Defaults to None.
            config (str | None, optional): Config file path. Defaults to None.
            config_env (str, optional): configure environment name. Defaults to 'default'.

        Raises:
            TypeError: `token`, `url`, or `config` is not str.
        """
        self.client = JijZeptClient(
            url=url, token=token, proxy=proxy, config=config, config_env=config_env
        )


@dataclasses.dataclass
class ParameterSearchParameters:
    """Data class for parameter_search_parameters."""

    multipliers: dict[str, float] = dataclasses.field(default_factory=lambda: {})
    mul_search: bool = False
    num_search: int = 15
    algorithm: str | None = None
    normalize_qubo: bool = True


def sample_model(
    client: JijZeptClient,
    solver: str,
    queue_name: str,
    problem: Problem,
    instance_data: dict[str, list | np.ndarray],
    fixed_variables: dict[str, dict[tuple[int, ...], int]],
    parameter_search_parameters: ParameterSearchParameters,
    max_wait_time: int | float | None,
    sync: bool,
    **kwargs,
) -> JijModelingResponse:
    """
    Solver using the given parameters, returns a JijModelingResponse object.

    Args:
        client (JijZeptClient): JijZept client object.
        solver (str): Solver type.
        queue_name (str): Queue name to use.
        problem (Problem): The problem to be solved.
        instance_data (dict[str, list | np.ndarray]): Instance data.
        fixed_variables (dict[str, dict[tuple[int, ...], int]]): Fixed variables.
        parameter_search_parameters (ParameterSearchParameters): Parameters for parameter search.
        max_wait_time (int | float | None): Maximum time to wait for the response, in seconds.
        sync (bool): Whether to wait for the response.

    Returns:
        JijModelingResponse: Stores minimum energy samples and statistics of the computation.

    Raises:
        JijZeptSolvingFailedError: If the computation failed to solve the problem.
        JijZeptSolvingUnknownError: If the computation ended in an unknown error.
        JijZeptSolvingValidationError: If the computation failed to validate the input parameters.
    """
    start = time.time()

    instance = {
        "mathematical_model": CEscape(jm.to_protobuf(problem), as_utf8=False),
        "instance_data": InstanceTranslator.instance_translate(instance_data),
    }

    if "timeout" in kwargs:
        warnings.warn(
            "The 'timeout' argument is deprecated and will be removed in a future. Please use 'max_wait_time' instead.",
            FutureWarning,
        )
        max_wait_time = kwargs.pop("timeout")

    param_dict = dataclasses.asdict(parameter_search_parameters)
    param_dict["fixed_variables"] = serialize_fixed_var(fixed_variables)
    param_dict.update(kwargs)

    response = post_instance_and_query(
        JijModelingResponse,
        client=client,
        instance_type="JijModeling",
        instance=instance,
        queue_name=queue_name,
        solver=solver,
        parameters=param_dict,
        timeout=max_wait_time,
        sync=sync,
    )
    response.sample_set.measuring_time = jm.MeasuringTime(
        solve=response.sample_set.measuring_time.solve,
        system=response.sample_set.measuring_time.system,
        total=time.time() - start,
    )

    # Raise error if the problem is not solved.
    if response.status == APIStatus.FAILED:
        raise JijZeptSolvingFailedError(
            response.error_message.get("message", "The problem is not solved.")
        )
    elif response.status == APIStatus.UNKNOWNERROR:
        raise JijZeptSolvingUnknownError(
            response.error_message.get("message", "The problem is not solved.")
        )
    elif response.status == APIStatus.VALIDATIONERROR:
        raise JijZeptSolvingValidationError(
            response.error_message.get("message", "The problem is not solved.")
        )

    return response


def merge_params_and_kwargs(
    parameters: typ.Type | None, kwargs: dict[str, typ.Any], param_cls: typ.Type
) -> dict[str, typ.Any]:
    """
    Merge the set of parameters defined by `parameters` and the set of parameters
    defined by `kwargs`, giving precedence to the values defined in `kwargs`.

    Args:
        parameters: A type or None representing the class defining the set of parameters.
        kwargs: A dictionary containing the values of the parameters to be used in a computation.
        param_cls: A type representing the class to be used for `parameters`.

    Returns:
        A dictionary containing the merged set of parameters and keyword arguments.
    """
    from jijzept.sampler import thirdparty

    param_dict: dict[str, typ.Any] = {}
    if parameters is None:
        parameters = dataclasses.asdict(param_cls())
    else:
        parameters = dataclasses.asdict(parameters)

    if param_cls.__name__ in thirdparty.__all__:
        param_dict.update({"parameters": parameters})
    else:
        param_dict.update(parameters)

    for key, value in kwargs.items():
        param_dict[key] = value
    return param_dict


def check_kwargs_against_dataclass(
    kwargs: dict[str, typ.Any],
    param_cls: typ.Type,
) -> None:
    """
    Check if the keyword arguments have a key other than the field names of specified dataclass.

    Args:
        kwargs: A dictionary of keyword arguments to be checked.
        param_cls: A type representing the class to be used for `parameters`.

    Raises:
        JijZeptClientValidationError: If **kwargs contain extra keyword argument.
    """
    field_names = dataclasses.asdict(param_cls()).keys()
    extra_kwargs: list[str] = []

    for key in kwargs.keys():
        if key == "timeout": continue  # "timeout"非対応化時、この行を削除してください
        if key not in field_names:
            extra_kwargs.append(f"`{key}`")

    if len(extra_kwargs) > 0:
        raise JijZeptClientValidationError(
            f"Extra keyword argument is permmited: {','.join(extra_kwargs)}. "
            f"**kwargs are allowed only for field names of {param_cls.__name__}."
        )
