from __future__ import annotations

import dataclasses

from numbers import Number

import numpy as np

from jijmodeling import Problem

from jijzept.entity.schema import SolverType
from jijzept.response import JijModelingResponse
from jijzept.sampler.base_sampler import (
    JijZeptBaseSampler,
    ParameterSearchParameters,
    check_kwargs_against_dataclass,
    merge_params_and_kwargs,
    sample_model,
)


@dataclasses.dataclass
class JijSXAuroraParameters:
    """Parameters for Jij SX-Aurora Annealing.

    Attributes:
        num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, 1000 will be set.
        beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
        beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
        num_reads (Optional[int], optional): The number of samples. If `None`, this will be set automatically.
        tsallis_q (Optional[float], optional): q value of tsallis statistics. If `None`, this will be set automatically.
        needs_square_constraints (Optional[dict[str, bool]], optional): This dictionary object is utilized to determine whether to square the constraint condition while incorporating it into the QUBO/HUBO penalty term. Here, the constraint's name is used as the key. If the value is set to True, the corresponding constraint is squared upon its addition to the QUBO/HUBO penalty term. By default, the value is set to True for linear constraints, and to False for non-linear ones.
        relax_as_penalties (Optional[dict[str, bool]], optional): This dictionary object is designed to regulate the incorporation of constraint conditions into the QUBO/HUBO penalty term, with the constraint's name functioning as the key. If the key's value is True, the respective constraint is added to the QUBO/HUBO penalty term. If the value is False, the constraint is excluded from the penalty term, though it remains subject to evaluation to verify if it meets the constraint conditions. By default, all constraint conditions have this value set to True.
    """

    num_sweeps: int | None = None
    beta_min: float | None = None
    beta_max: float | None = None
    num_reads: int | None = None
    tsallis_q: float | None = None
    needs_square_constraints: dict[str, bool] | None = None
    relax_as_penalties: dict[str, bool] | None = None


class JijSXAuroraSampler(JijZeptBaseSampler):
    """Sampler using the Jij SX-Aurora Annealing."""

    jijmodeling_solver_type = SolverType(
        queue_name="sxaurorasolver", solver="SAParaSearch"
    )

    def sample_model(
        self,
        problem: Problem,
        feed_dict: dict[str, Number | list | np.ndarray],
        multipliers: dict[str, Number] = {},
        fixed_variables: dict[str, dict[tuple[int, ...], int | float]] = {},
        search: bool = False,
        num_search: int = 15,
        algorithm: str | None = None,
        parameters: JijSXAuroraParameters | None = None,
        max_wait_time: int | float | None = None,
        sync: bool = True,
        queue_name: str | None = None,
        **kwargs,
    ) -> JijModelingResponse:
        """Sample using JijModeling by means of Jij SX-Aurora Annealing solver.

        To configure the solver, instantiate the `JijSXAuroraParameters` class and pass the instance to the `parameters` argument.

        Args:
            problem (Problem): Mathematical expression of JijModeling.
            feed_dict (Dict[str, Union[Number, list, np.ndarray]]): The actual values to be assigned to the placeholders.
            multipliers (Dict[str, Number], optional): The actual multipliers for penalty terms, derived from constraint conditions.
            fixed_variables (Dict[str, Dict[Tuple[int, ...], Union[int, float]]]): dictionary of variables to fix.
            search (bool, optional): If `True`, the parameter search will be carried out, which tries to find better values of multipliers for penalty terms.
            num_search (int, optional): The number of parameter search iteration. Defaults to set 15. This option works if `search` is `True`.
            algorithm (Optional[str], optional): Algorithm for parameter search. Defaults to None.
            parameters (Optional[JijSXAuroraParameters], optional): parameters for SXAurora. Defaults to None.
            max_wait_time (int | float | None, optional): The number of timeout [sec] for post request. If `None`, 60 (one minute) will be set. Please note that this argument is for the `jijzept` timeout and not for configuring solver settings, such as solving time.
            sync (bool, optional): Synchronous mode.
            queue_name (Optional[str], optional): Queue name.
            **kwargs: SXAurora parameters using **kwargs. If both `**kwargs` and `parameters` are exist, the value of `**kwargs` takes precedence.

        Returns:
            JijModelingResponse: Stores minimum energy samples and other information.

        Examples:
            ```python
            import jijzept as jz
            import jijmodeling as jm
            n = jm.Placeholder('n')
            x = jm.Binary('x', shape=n)
            d = jm.Placeholder('d', shape=n)
            i = jm.Element("i", n)
            problem = jm.Problem('problem')
            problem += jm.Sum(i, d[i] * x[i])
            sampler = jz.JijSXAuroraSampler(config='config.toml')
            response = sampler.sample_model(problem, feed_dict={'n': 5, 'd': [1,2,3,4,5]})
            ```
        """
        check_kwargs_against_dataclass(kwargs, JijSXAuroraParameters)
        param_dict = merge_params_and_kwargs(parameters, kwargs, JijSXAuroraParameters)

        para_search_params = ParameterSearchParameters(
            multipliers=multipliers,
            mul_search=search,
            num_search=num_search,
            algorithm=algorithm,
        )

        if queue_name is None:
            queue_name = self.jijmodeling_solver_type.queue_name

        sample_set = sample_model(
            self.client,
            self.jijmodeling_solver_type.solver,
            queue_name=queue_name,
            problem=problem,
            instance_data=feed_dict,
            fixed_variables=fixed_variables,
            parameter_search_parameters=para_search_params,
            max_wait_time=max_wait_time,
            sync=sync,
            **param_dict,
        )
        return sample_set
